;(function($B){

var _b_ = $B.builtins

var object = _b_.object

function $err(op, other){
    var msg = "unsupported operand type(s) for " + op +
        ": 'float' and '" + $B.class_name(other) + "'"
    throw _b_.TypeError.$factory(msg)
}

function float_value(obj){
    // Instances of float subclasses that call float.__new__(cls, value)
    // have an attribute $brython_value set
    return obj.$brython_value !== undefined ? obj.$brython_value : obj
}

// dictionary for built-in class 'float'
var float = {
    __class__: _b_.type,
    __dir__: object.__dir__,
    $infos: {
        __module__: "builtins",
        __name__: "float"
    },
    $is_class: true,
    $native: true,
    $descriptors: {
        "numerator": true,
        "denominator": true,
        "imag": true,
        "real": true
    }
}

float.numerator = function(self){return float_value(self)}
float.denominator = function(self){return _b_.int.$factory(1)}
float.imag = function(self){return _b_.int.$factory(0)}
float.real = function(self){return float_value(self)}
float.__float__ = function(self){return float_value(self)}

// cache lshifts of 1
$B.shift1_cache = {}

float.as_integer_ratio = function(self){
    self = float_value(self)

    if(self.valueOf() == Number.POSITIVE_INFINITY ||
            self.valueOf() == Number.NEGATIVE_INFINITY){
        throw _b_.OverflowError.$factory("Cannot pass infinity to " +
            "float.as_integer_ratio.")
    }
    if(! Number.isFinite(self.valueOf())){
        throw _b_.ValueError.$factory("Cannot pass NaN to " +
            "float.as_integer_ratio.")
    }

    var tmp = frexp(self.valueOf()),
        fp = tmp[0],
        exponent = tmp[1]

    for (var i = 0; i < 300; i++){
        if(fp == Math.floor(fp)){
            break
        }else{
            fp *= 2
            exponent--
        }
    }

    numerator = _b_.int.$factory(fp)
    py_exponent = _b_.abs(exponent)
    denominator = 1
    var x
    if($B.shift1_cache[py_exponent] !== undefined){
        x = $B.shift1_cache[py_exponent]
    }else{
        x = $B.$getattr(1, "__lshift__")(py_exponent)
        $B.shift1_cache[py_exponent] = x
    }
    py_exponent = x
    if(exponent > 0){
        numerator = $B.rich_op("__mul__", numerator, py_exponent)
    }else{
        denominator = py_exponent
    }

    return $B.fast_tuple([_b_.int.$factory(numerator),
        _b_.int.$factory(denominator)])
}

float.__abs__ = function(self){
    return new Number(Math.abs(float_value(self)))
}

float.__bool__ = function(self){
    self = float_value(self)
    return _b_.bool.$factory(self.valueOf())
}

float.__divmod__ = function(self, other){
    if(! _b_.isinstance(other, [_b_.int, float])){
        return _b_.NotImplemented
    }
    return $B.fast_tuple([float.__floordiv__(self, other),
        float.__mod__(self, other)])
}

float.__eq__ = function(self, other){
    self = float_value(self)
    other = float_value(other)
    if(isNaN(self) && isNaN(other)){return false}
    if(_b_.isinstance(other, _b_.int)){return self == other}
    if(_b_.isinstance(other, float)) {
      // new Number(1.2) == new Number(1.2) returns false !!!
      return self.valueOf() == other.valueOf()
    }
    if(_b_.isinstance(other, _b_.complex)){
      if (other.$imag != 0){return false}
      return self == other.$real
    }
    return _b_.NotImplemented
}

float.__floordiv__ = function(self, other){
    self = float_value(self)
    other = float_value(other)
    if(_b_.isinstance(other,[_b_.int, float])){
      if(other.valueOf() == 0){
          throw _b_.ZeroDivisionError.$factory('division by zero')
      }
      return float.$factory(Math.floor(self / other))
    }
    return _b_.NotImplemented
}

float.fromhex = function(arg){
   // [sign] ['0x'] integer ['.' fraction] ['p' exponent]

   if(! _b_.isinstance(arg, _b_.str)){
      throw _b_.ValueError.$factory("argument must be a string")
   }

   var value = arg.trim()   // remove leading and trailing whitespace
   switch(value.toLowerCase()) {
      case "+inf":
      case "inf":
      case "+infinity":
      case "infinity":
          return $FloatClass(Infinity)
      case "-inf":
      case "-infinity":
          return $FloatClass(-Infinity)
      case "+nan":
      case "nan":
          return $FloatClass(Number.NaN)
      case "-nan":
          return $FloatClass(-Number.NaN)
      case "":
          throw _b_.ValueError.$factory("could not convert string to float")
   }

   var mo = /^(\d*)(\.?)(\d*)$/.exec(value)

   if(mo !== null){
       var res = parseFloat(mo[1]),
           coef = 16
       if(mo[2]){
           for(var digit of mo[3]){
               res += parseInt(digit, 16) / coef
               coef *= 16
           }
       }
       return $FloatClass(res)
   }

   // lets see if this is a hex string.
   var _m = /^(\+|-)?(0x)?([0-9A-F]+\.?)?(\.[0-9A-F]+)?(p(\+|-)?\d+)?$/i.exec(value)

   if(_m == null){
       throw _b_.ValueError.$factory("invalid hexadecimal floating-point string")
   }

   var _sign = _m[1],
       _int = parseInt(_m[3] || '0', 16),
       _fraction = _m[4] || '.0',
       _exponent = _m[5] || 'p0'

   if(_sign == "-"){_sign = -1}else{_sign = 1}

   var _sum = _int

   for(var i = 1, len = _fraction.length; i < len; i++){
       _sum += parseInt(_fraction.charAt(i), 16) / Math.pow(16, i)
   }
   return new Number(_sign * _sum * Math.pow(2,
       parseInt(_exponent.substring(1))))
}

float.__getformat__ = function(arg){
    if(arg == "double" || arg == "float"){return "IEEE, little-endian"}
    throw _b_.ValueError.$factory("__getformat__() argument 1 must be " +
        "'double' or 'float'")
}

function preformat(self, fmt){
    if(fmt.empty){
        return _b_.str.$factory(self)
    }
    if(fmt.type && 'eEfFgGn%'.indexOf(fmt.type) == -1){
        throw _b_.ValueError.$factory("Unknown format code '" + fmt.type +
            "' for object of type 'float'")
    }
    if(isNaN(self)){
        if(fmt.type == "f" || fmt.type == "g"){return "nan"}
        else{return "NAN"}
    }
    if(self == Number.POSITIVE_INFINITY){
        if(fmt.type == "f" || fmt.type == "g"){return "inf"}
        else{return "INF"}
    }
    if(fmt.precision === undefined && fmt.type !== undefined){
        fmt.precision = 6
    }
    if(fmt.type == "%"){self *= 100}

    if(fmt.type == "e"){
        var res = self.toExponential(fmt.precision),
            exp = parseInt(res.substr(res.search("e") + 1))
            if(Math.abs(exp) < 10){
                res = res.substr(0, res.length - 1) + "0" +
                    res.charAt(res.length - 1)
            }
        return res
    }

    if(fmt.precision !== undefined){
        // Use Javascript toFixed to get the correct result
        // The argument of toFixed is the number of digits after .
        var prec = fmt.precision
        if(prec == 0){return Math.round(self) + ""}
        var res = self.toFixed(prec),
            pt_pos = res.indexOf(".")
        if(fmt.type !== undefined &&
                (fmt.type == "%" || fmt.type.toLowerCase() == "f")){
            if(pt_pos == -1){
                res += "." + "0".repeat(fmt.precision)
            }else{
                var missing = fmt.precision - res.length + pt_pos + 1
                if(missing > 0){
                    res += "0".repeat(missing)
                }
            }
        }else if(fmt.type && fmt.type.toLowerCase() == "g"){
            var exp_fmt = preformat(self, {type: "e"}).split("e"),
                exp = parseInt(exp_fmt[1])
            if(-4 <= exp && exp < fmt.precision){
                res = preformat(self,
                        {type: "f", precision: fmt.precision - 1 - exp})
            }else{
                res = preformat(self,
                    {type: "e", precision: fmt.precision - 1})
            }
            var parts = res.split("e")
            if(fmt.alternate){
                if(parts[0].search(/\./) == -1){
                    parts[0] += '.'
                }
            }else{
                if(parts[1]){
                    var signif = parts[0]
                    while(signif.endsWith("0")){
                        signif = signif.substr(0, signif.length - 1)
                    }
                    if(signif.endsWith(".")){
                        signif = signif.substr(0, signif.length - 1)
                    }
                    parts[0] = signif
                }
            }
            res = parts.join("e")
            if(fmt.type == "G"){
                res = res.toUpperCase()
            }
            return res
        }else if(fmt.type === undefined){
            fmt.type = "g"
            res = preformat(self, fmt)
            fmt.type = undefined
        }else{
            var res1 = self.toExponential(fmt.precision - 1),
                exp = parseInt(res1.substr(res1.search("e") + 1))
            if(exp < -4 || exp >= fmt.precision - 1){
                var elts = res1.split("e")
                // Remove trailing 0 from mantissa
                while(elts[0].endsWith("0")){elts[0] = elts[0].substr(0,
                    elts[0].length - 1)}
                res = elts.join("e")
            }
        }
    }else{var res = _b_.str.$factory(self)}

    if(fmt.type === undefined || "gGn".indexOf(fmt.type) != -1){
        // remove trailing 0 for non-exponential formats
        if(res.search("e") == -1){
            while(res.charAt(res.length - 1) == "0"){
                res = res.substr(0, res.length - 1)
            }
        }
        if(res.charAt(res.length - 1) == "."){
            if(fmt.type === undefined){res += "0"}
            else{res = res.substr(0, res.length - 1)}
        }
    }
    if(fmt.sign !== undefined){
        if((fmt.sign == " " || fmt.sign == "+" ) && self > 0){
            res = fmt.sign + res
        }
    }
    if(fmt.type == "%"){res += "%"}

    return res
}

float.__format__ = function(self, format_spec){
    self = float_value(self)
    var fmt = new $B.parse_format_spec(format_spec)
    fmt.align = fmt.align || ">"
    var raw = preformat(self, fmt).split('.'),
        _int = raw[0]
    if(fmt.comma){
        var len = _int.length, nb = Math.ceil(_int.length / 3), chunks = []
        for(var i = 0; i < nb; i++){
            chunks.push(_int.substring(len - 3 * i - 3, len - 3 * i))
        }
        chunks.reverse()
        raw[0] = chunks.join(",")
    }
    return $B.format_width(raw.join("."), fmt)
}

float.__hash__ = function(self) {
    if(self === undefined){
       return float.__hashvalue__ || $B.$py_next_hash--  // for hash of float type (not instance of int)
    }

    var _v = self.valueOf()
    if(_v === Infinity){return 314159}
    if(_v === -Infinity){return -271828}
    if(isNaN(_v)){return 0}
    // for integers, return the value
    if(_v == Math.round(_v)){return Math.round(_v)}

    var r = frexp(_v)
    r[0] *= Math.pow(2, 31)
    var hipart = _b_.int.$factory(r[0])
    r[0] = (r[0] - hipart) * Math.pow(2, 31)
    var x = hipart + _b_.int.$factory(r[0]) + (r[1] << 15)
    return x & 0xFFFFFFFF
}

function isninf(x) {
    var x1 = x
    if(_b_.isinstance(x, float)){x1 = float.numerator(x)}
    return x1 == -Infinity || x1 == Number.NEGATIVE_INFINITY
}

function isinf(x) {
    var x1 = x
    if((! x instanceof Number) && _b_.isinstance(x, float)){
        x1 = float.numerator(x)
    }
    return x1 == Infinity || x1 == -Infinity ||
        x1 == Number.POSITIVE_INFINITY || x1 == Number.NEGATIVE_INFINITY
}

function isnan(x) {
    var x1 = x
    if(_b_.isinstance(x, float)){x1 = float.numerator(x)}
    return isNaN(x1)
}

function fabs(x){
    if(x == 0){
        return new Number(0)
    }
    return x > 0 ? float.$factory(x) : float.$factory(-x)
}

function frexp(x){
    var x1 = x
    if(_b_.isinstance(x, float)){
        x1 = x.valueOf()
    }

    if(isNaN(x1) || isinf(x1)){
        return [x1, -1]
    }else if (x1 == 0){
        return [0, 0]
    }

    var sign = 1,
        ex = 0,
        man = x1

    if(man < 0.){
       sign = -sign
       man = -man
    }

    while(man < 0.5){
       man *= 2.0
       ex--
    }

    while(man >= 1.0){
       man *= 0.5
       ex++
    }

    man *= sign

    return [man, ex]
}

function ldexp(x, i) {
    if(isninf(x)){return float.$factory('-inf')}
    if(isinf(x)){return float.$factory('inf')}

    var y = x
    if(_b_.isinstance(x, float)){y = x.valueOf()}
    if(y == 0){return y}

    var j = i
    if(_b_.isinstance(i, float)){j = i.valueOf()}
    return y * Math.pow(2, j)
}

float.$funcs = {isinf, isninf, isnan, fabs, frexp, ldexp}

float.hex = function(self) {
    // http://hg.python.org/cpython/file/d422062d7d36/Objects/floatobject.c
    self = float_value(self)
    var DBL_MANT_DIG = 53,   // 53 bits?
        TOHEX_NBITS = DBL_MANT_DIG + 3 - (DBL_MANT_DIG + 2) % 4

    switch(self.valueOf()) {
        case Infinity:
        case -Infinity:
        case Number.NaN:
        case -Number.NaN:
            return self
        case -0:
            return "-0x0.0p0"
        case 0:
            return "0x0.0p0"
    }

    var _a = frexp(fabs(self.valueOf())),
        _m = _a[0],
        _e = _a[1],
        _shift = 1 - Math.max(-1021 - _e, 0)

    _m = ldexp(_m, _shift)
    _e -= _shift

    var _int2hex = "0123456789ABCDEF".split(""),
        _s = _int2hex[Math.floor(_m)]
    _s += '.'
    _m -= Math.floor(_m)

    for(var i = 0; i < (TOHEX_NBITS - 1) / 4; i++){
        _m *= 16.0
        _s += _int2hex[Math.floor(_m)]
        _m -= Math.floor(_m)
    }

    var _esign = "+"
    if(_e < 0){
       _esign = "-"
       _e = -_e
    }

    if(self.value < 0){return "-0x" + _s + "p" + _esign + _e}
    return "0x" + _s + "p" + _esign + _e
}

float.__init__ = function(self, value){
    return _b_.None
}

float.__int__ = function(self){return parseInt(self)}

float.is_integer = function(self) {return _b_.int.$factory(self) == self}

float.__mod__ = function(self, other) {
    // can't use Javascript % because it works differently for negative numbers
    self = float_value(self)
    other = float_value(other)
    if(other == 0){
        throw _b_.ZeroDivisionError.$factory("float modulo")
    }
    if(_b_.isinstance(other, _b_.int)){
        other = _b_.int.numerator(other)
        return new Number((self % other + other) % other)
    }

    if(_b_.isinstance(other, float)){
        // use truncated division
        // cf https://en.wikipedia.org/wiki/Modulo_operation
        var q = Math.floor(self / other),
            r = self - other * q
        return new Number(r)
    }
    return _b_.NotImplemented
}

float.__mro__ = [object]

float.__mul__ = function(self, other){
    self = float_value(self)
    other = float_value(other)
    if(_b_.isinstance(other, _b_.int)){
        if(other.__class__ == $B.long_int){
            return new Number(self * parseFloat(other.value))
        }
        other = _b_.int.numerator(other)
        return new Number(self * other)
    }
    if(_b_.isinstance(other, float)){
        return new Number(self * float_value(other))
    }
    return _b_.NotImplemented
}

float.__ne__ = function(self, other){
    var res = float.__eq__(self, other)
    return res === _b_.NotImplemented ? res : ! res
}

float.__neg__ = function(self){
    return new Number(-float_value(self))
}

float.__new__ = function(cls, value){
    if(cls === undefined){
        throw _b_.TypeError.$factory("float.__new__(): not enough arguments")
    }else if(! _b_.isinstance(cls, _b_.type)){
        throw _b_.TypeError.$factory("float.__new__(X): X is not a type object")
    }
    if(cls === float){return float.$factory(value)}
    return {
        __class__: cls,
        __dict__: $B.empty_dict(),
        $brython_value: value || 0
    }
}

float.__pos__ = function(self){return float_value(self)}

float.__pow__ = function(self, other){
    self = float_value(self)
    other = float_value(other)

    var other_int = _b_.isinstance(other, _b_.int)
    if(other_int || _b_.isinstance(other, float)){
        if(self == 1){return new Number(1)} // even for Infinity or NaN
        if(other == 0){return new Number(1)}

        if(self == -1 &&
                (! isFinite(other) || other.__class__ === $B.long_int ||
                     ! $B.is_safe_int(other)) &&
                ! isNaN(other)){
            return new Number(1)
        }else if(self == 0 && isFinite(other) && other < 0){
            throw _b_.ZeroDivisionError.$factory("0.0 cannot be raised " +
              "to a negative power")
        }else if(self == Number.NEGATIVE_INFINITY && ! isNaN(other)){
            if(other < 0 && other % 2 == 1){
                return new Number(-0.0)
            }else if(other < 0){return new Number(0)}
            else if(other > 0 && other % 2 == 1){
                return Number.NEGATIVE_INFINITY
            }else{return Number.POSITIVE_INFINITY}
        }else if(self == Number.POSITIVE_INFINITY && ! isNaN(other)){
            return other > 0 ? self : new Number(0)
        }
        if(other == Number.NEGATIVE_INFINITY && ! isNaN(self)){
            return Math.abs(self) < 1 ? Number.POSITIVE_INFINITY :
                new Number(0)
        }else if(other == Number.POSITIVE_INFINITY  && ! isNaN(self)){
            return Math.abs(self) < 1 ? new Number(0) :
                Number.POSITIVE_INFINITY
        }
        if(self < 0 &&
                ! $B.$getattr(other, "__eq__")(_b_.int.$factory(other))){
            // use complex power
            return _b_.complex.__pow__($B.make_complex(self, 0), other)
        }
        return float.$factory(Math.pow(self, other))
    }
    return _b_.NotImplemented
}

function __newobj__(){
    // __newobj__ is called with a generator as only argument
    var $ = $B.args('__newobj__', 0, {}, [], arguments, {}, 'args', null),
        args = $.args
    var res = args.slice(1)
    res.__class__ = args[0]
    return res
}

float.__reduce_ex__ = function(self){
    return $B.fast_tuple([
        __newobj__,
        $B.fast_tuple([self.__class__ || _b_.int, float_value(self)]),
        _b_.None,
        _b_.None,
        _b_.None])
}

float.__repr__ = function(self){
    $B.builtins_repr_check(float, arguments) // in brython_builtins.js
    self = float_value(self).valueOf()
    if(self == Infinity){
        return 'inf'
    }else if(self == -Infinity){
        return '-inf'
    }else if(isNaN(self)){
        return 'nan'
    }else if(self === 0){
        if(1 / self === -Infinity){
            return '-0.0'
        }
        return '0.0'
    }

    var res = self + "" // coerce to string

    if(res.search(/[.eE]/) == -1){
        res += ".0"
    }
    var split_e = res.split(/e/i)
    if(split_e.length == 2){
        var mant = split_e[0],
            exp = split_e[1]
        if(exp.startsWith('-')){
            exp_str = parseInt(exp.substr(1)) + ''
            if(exp_str.length < 2){
                exp_str = '0' + exp_str
            }
            return mant + 'e-' + exp_str
        }
    }
    var x, y
    [x, y] = res.split('.')
    var sign = ''
    if(x[0] == '-'){
        x = x.substr(1)
        sign = '-'
    }
    if(x.length > 16){
        var exp = x.length - 1,
            int_part = x[0],
            dec_part = x.substr(1) + y
        while(dec_part.endsWith("0")){
            dec_part = dec_part.substr(0, dec_part.length - 1)
        }
        var mant = int_part
        if(dec_part.length > 0){
            mant += '.' + dec_part
        }
        return sign + mant + 'e+' + exp
    }else if(x == "0"){
        var exp = 0
        while(exp < y.length && y.charAt(exp) == "0"){
            exp++
        }
        if(exp > 3){
            // form 0.0000xyz
            var rest = y.substr(exp),
                exp = (exp + 1).toString()
            while(rest.endsWith("0")){
                rest = rest.substr(0, res.length - 1)
            }
            var mant = rest[0]
            if(rest.length > 1){
                mant += '.' + rest.substr(1)
            }
            if(exp.length == 1){
                exp = '0' + exp
            }
            return sign + mant + 'e-' + exp
        }
    }
    return _b_.str.$factory(res)
}

float.__setattr__ = function(self, attr, value){
    if(self.constructor === Number){
        if(float[attr] === undefined){
            throw _b_.AttributeError.$factory("'float' object has no attribute '" +
                attr + "'")
        }else{
            throw _b_.AttributeError.$factory("'float' object attribute '" +
                attr + "' is read-only")
        }
    }
    // subclasses of float can have attributes set
    self[attr] = value
    return _b_.None
}

float.__truediv__ = function(self, other){
    self = float_value(self)
    other = float_value(other)
    if(_b_.isinstance(other, [_b_.int, float])){
        if(other.valueOf() == 0){
            throw _b_.ZeroDivisionError.$factory("division by zero")
        }
        return float.$factory(self/other)
    }
    return _b_.NotImplemented
}

// operations
var $op_func = function(self, other){
    self = float_value(self)
    other = float_value(other)
    if(_b_.isinstance(other, _b_.int)){
        if(typeof other == "boolean"){
            return other ? self - 1 : self
        }else if(other.__class__ === $B.long_int){
            return float.$factory(self - parseInt(other.value))
        }else{return float.$factory(self - other)}
    }
    if(_b_.isinstance(other, float)){
        return float.$factory(self - other)
    }
    return _b_.NotImplemented
}
$op_func += "" // source code
var $ops = {"+": "add", "-": "sub"}
for(var $op in $ops){
    var $opf = $op_func.replace(/-/gm, $op)
    $opf = $opf.replace(/__rsub__/gm, "__r" + $ops[$op] + "__")
    eval("float.__" + $ops[$op] + "__ = " + $opf)
}

// comparison methods
var $comp_func = function(self, other){
    self = float_value(self)
    other = float_value(other)

    if(_b_.isinstance(other, _b_.int)){
        if(other.__class__ === $B.long_int){
            return self > parseInt(other.value)
        }
        return self > other.valueOf()
    }
    if(_b_.isinstance(other,float)){return self > other}

    if(_b_.isinstance(other, _b_.bool)) {
      return self.valueOf() > _b_.bool.__hash__(other)
    }
    if(_b_.hasattr(other, "__int__") || _b_.hasattr(other, "__index__")) {
       return _b_.int.__gt__(self, $B.$GetInt(other))
    }

    // See if other has the opposite operator, eg <= for >
    var inv_op = $B.$getattr(other, "__le__", _b_.None)
    if(inv_op !== _b_.None){return inv_op(self)}

    throw _b_.TypeError.$factory(
        "unorderable types: float() > " + $B.class_name(other) + "()")
}

$comp_func += "" // source code
for(var $op in $B.$comps){
    eval("float.__" + $B.$comps[$op] + "__ = "+
          $comp_func.replace(/>/gm, $op).
              replace(/__gt__/gm, "__" + $B.$comps[$op] + "__").
              replace(/__le__/, "__" + $B.$inv_comps[$op] + "__"))
}

// add "reflected" methods
var r_opnames = ["add", "sub", "mul", "truediv", "floordiv", "mod", "pow",
    "lshift", "rshift", "and", "xor", "or", "divmod"]

for(var r_opname of r_opnames){
    if(float["__r" + r_opname + "__"] === undefined &&
            float['__' + r_opname + '__']){
        float["__r" + r_opname + "__"] = (function(name){
            return function(self, other){
                if(_b_.isinstance(other, _b_.int)){
                    other = float_value(_b_.int.numerator(other))
                    return float["__" + name + "__"](other, self)
                }else if(_b_.isinstance(other, float)){
                    other = float_value(other)
                    return float["__" + name + "__"](other, self)
                }
                return _b_.NotImplemented
            }
        })(r_opname)
    }
}

function $FloatClass(value){
    return new Number(value)
}

function to_digits(s){
    // Transform a string to another string where all arabic-indic digits
    // are converted to latin digits
    var arabic_digits = "\u0660\u0661\u0662\u0663\u0664\u0665\u0666\u0667\u0668\u0669",
        res = ""
    for(var i = 0; i < s.length; i++){
        var x = arabic_digits.indexOf(s[i])
        if(x > -1){res += x}
        else{res += s[i]}
    }
    return res
}

// constructor for built-in class 'float'
float.$factory = function (value){
    switch(value) {
        case undefined:
            return $FloatClass(0.0)
        case Number.MAX_VALUE:
            //take care of 'inf not identical to 1.797...e+308' error
            return $FloatClass(Infinity)
        case -Number.MAX_VALUE:
            return $FloatClass(-Infinity)
        case true:
            return new Number(1)
        case false:
            return new Number(0)
    }

    if(typeof value == "number"){
        return new Number(value)
    }
    if(_b_.isinstance(value, float)){
        return float_value(value)
    }
    if(_b_.isinstance(value, _b_.bytes)){
      var s = $B.$getattr(value, "decode")("latin-1")
      return float.$factory($B.$getattr(value, "decode")("latin-1"))
    }

    if(typeof value == "string"){
       value = value.trim()   // remove leading and trailing whitespace
       switch(value.toLowerCase()) {
           case "+inf":
           case "inf":
           case "+infinity":
           case "infinity":
               return Number.POSITIVE_INFINITY
           case "-inf":
           case "-infinity":
               return Number.NEGATIVE_INFINITY
           case "+nan":
           case "nan":
               return Number.NaN
           case "-nan":
               return -Number.NaN
           case "":
               throw _b_.ValueError.$factory("count not convert string to float")
           default:
               value = value.charAt(0) + value.substr(1).replace(/_/g, "") // PEP 515
               value = to_digits(value) // convert arabic-indic digits to latin
               if (isFinite(value)) return $FloatClass(eval(value))
               else {
                   _b_.str.encode(value, "latin-1") // raises UnicodeEncodeError if not valid
                   throw _b_.ValueError.$factory(
                       "Could not convert to float(): '" +
                       _b_.str.$factory(value) + "'")
               }
         }
    }
    var klass = value.__class__ || $B.get_class(value),
        num_value = $B.to_num(value, ["__float__", "__index__"])

    if(value !== Number.POSITIVE_INFINITY && ! isFinite(num_value)){
        throw _b_.OverflowError.$factory('int too large to convert to float')
    }
    if(num_value !== null){
        return num_value
    }

    throw _b_.TypeError.$factory("float() argument must be a string or a " +
        "number, not '" + $B.class_name(value) + "'")
}

$B.$FloatClass = $FloatClass

$B.set_func_names(float, "builtins")

// Dictionary and factory for subclasses of float
var FloatSubclass = $B.FloatSubclass  = {
    __class__: _b_.type,
    __mro__: [object],
    $infos: {
        __module__: "builtins",
        __name__: "float"
    },
    $is_class: true
}

for(var $attr in float){
    if(typeof float[$attr] == "function"){
        FloatSubclass[$attr] = (function(attr){
            return function(){
                var args = [], pos = 0
                if(arguments.length > 0){
                    var args = [arguments[0].valueOf()], pos = 1
                    for(var i = 1, len = arguments.length; i < len; i++){
                        args[pos++] = arguments[i]
                    }
                }
                return float[attr].apply(null, args)
            }
        })($attr)
    }
}

$B.set_func_names(FloatSubclass, "builtins")

_b_.float = float
})(__BRYTHON__)
